package xyz.lychee.lagfixer.nms.v1_21_R2;

import com.google.common.collect.ImmutableSet;
import java.util.Comparator;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.List;
import java.util.Objects;
import java.util.stream.Stream;
import net.minecraft.server.level.WorldServer;
import net.minecraft.tags.TagKey;
import net.minecraft.tags.TagsItem;
import net.minecraft.world.entity.Entity;
import net.minecraft.world.entity.EntityCreature;
import net.minecraft.world.entity.EntityLiving;
import net.minecraft.world.entity.ai.attributes.GenericAttributes;
import net.minecraft.world.entity.ai.goal.PathfinderGoal;
import net.minecraft.world.entity.ai.goal.PathfinderGoalBreed;
import net.minecraft.world.entity.ai.goal.PathfinderGoalPanic;
import net.minecraft.world.entity.ai.goal.PathfinderGoalSelector;
import net.minecraft.world.entity.ai.goal.PathfinderGoalTempt;
import net.minecraft.world.entity.ai.goal.PathfinderGoal.Type;
import net.minecraft.world.entity.ai.targeting.PathfinderTargetCondition;
import net.minecraft.world.entity.ai.util.DefaultRandomPos;
import net.minecraft.world.entity.animal.EntityAnimal;
import net.minecraft.world.entity.animal.EntityBee;
import net.minecraft.world.entity.animal.EntityChicken;
import net.minecraft.world.entity.animal.EntityCow;
import net.minecraft.world.entity.animal.EntityFox;
import net.minecraft.world.entity.animal.EntityOcelot;
import net.minecraft.world.entity.animal.EntityPanda;
import net.minecraft.world.entity.animal.EntityParrot;
import net.minecraft.world.entity.animal.EntityPig;
import net.minecraft.world.entity.animal.EntityRabbit;
import net.minecraft.world.entity.animal.EntitySheep;
import net.minecraft.world.entity.animal.EntityTurtle;
import net.minecraft.world.entity.animal.EntityWolf;
import net.minecraft.world.entity.animal.armadillo.Armadillo;
import net.minecraft.world.entity.animal.axolotl.Axolotl;
import net.minecraft.world.entity.animal.camel.Camel;
import net.minecraft.world.entity.animal.frog.Frog;
import net.minecraft.world.entity.animal.goat.Goat;
import net.minecraft.world.entity.animal.horse.EntityHorse;
import net.minecraft.world.entity.animal.horse.EntityLlama;
import net.minecraft.world.entity.animal.sniffer.Sniffer;
import net.minecraft.world.entity.monster.EntityStrider;
import net.minecraft.world.entity.monster.hoglin.EntityHoglin;
import net.minecraft.world.entity.monster.piglin.EntityPiglin;
import net.minecraft.world.entity.player.EntityHuman;
import net.minecraft.world.item.Item;
import net.minecraft.world.phys.Vec3D;
import org.apache.commons.lang3.StringUtils;
import org.bukkit.configuration.ConfigurationSection;
import org.bukkit.craftbukkit.v1_21_R2.entity.CraftCreature;
import org.bukkit.craftbukkit.v1_21_R2.event.CraftEventFactory;
import org.bukkit.event.entity.EntityTargetLivingEntityEvent;
import org.bukkit.event.entity.EntityTargetEvent.TargetReason;
import xyz.lychee.lagfixer.modules.CustomAiModule;

public class CustomAi extends CustomAiModule.NMS {
   private static final PathfinderGoal optimized = new PathfinderGoal() {
      public boolean b() {
         return false;
      }
   };
   private static final HashMap<Class<? extends Entity>, TagKey<Item>> items = new HashMap();
   private final boolean collides;
   private final boolean silent;
   private final boolean mobchip;
   private final boolean typeAenabled;
   private final boolean typeBenabled;
   private final boolean typeCenabled;
   private final boolean temptEnabled;
   private final boolean breedEnabled;
   private final boolean panicEnabled;
   private final ImmutableSet<String> blacklistSet;
   private final ImmutableSet<String> whitelistSet;
   private final PathfinderTargetCondition temptTargeting;
   private final PathfinderTargetCondition breedTargeting;
   private final double temptSpeed;
   private final double breedSpeed;
   private final double panicSpeed;
   private final double panicRange;
   private final int temptCooldown;
   private final int panicCooldown;
   private final boolean temptTriggerBothHands;
   private final boolean temptEvent;
   private final boolean temptTeleport;
   private final boolean breedEvent;
   private final boolean breedTeleport;

   public CustomAi(CustomAiModule module) {
      super(module);
      ConfigurationSection section = this.getModule().getSection();
      this.collides = section.getBoolean("collides");
      this.silent = section.getBoolean("silent");
      this.mobchip = section.getBoolean("pathfinder.mobchip");
      this.typeAenabled = section.getBoolean("pathfinder.typeA.enabled");
      this.typeBenabled = section.getBoolean("pathfinder.typeB.enabled");
      this.typeCenabled = section.getBoolean("pathfinder.typeC.enabled");
      this.whitelistSet = ImmutableSet.copyOf(section.getStringList("pathfinder.typeA.whitelist"));
      this.blacklistSet = ImmutableSet.copyOf(section.getStringList("pathfinder.typeB.blacklist"));
      this.temptEnabled = section.getBoolean("animals.tempt.enabled");
      this.temptTargeting = this.targetingConditions(section, "animals.tempt");
      this.temptSpeed = section.getDouble("animals.tempt.speed");
      this.temptCooldown = section.getInt("animals.tempt.cooldown");
      this.temptTriggerBothHands = section.getBoolean("animals.tempt.trigger_two_hands");
      this.temptEvent = section.getBoolean("animals.tempt.event");
      this.temptTeleport = section.getBoolean("animals.tempt.teleport");
      this.breedEnabled = section.getBoolean("animals.tempt.enabled");
      this.breedTargeting = this.targetingConditions(section, "animals.breed");
      this.breedSpeed = section.getDouble("animals.breed.speed");
      this.breedEvent = section.getBoolean("animals.breed.event");
      this.breedTeleport = section.getBoolean("animals.breed.teleport");
      this.panicEnabled = section.getBoolean("animals.tempt.enabled");
      this.panicCooldown = section.getInt("animals.panic.cooldown");
      this.panicSpeed = section.getDouble("animals.panic.speed");
      this.panicRange = Math.pow(section.getDouble("animals.panic.range"), 2.0);
   }

   private PathfinderTargetCondition targetingConditions(ConfigurationSection section, String key) {
      return PathfinderTargetCondition.b().a(section.getDouble(key + ".range")).d();
   }

   public void setCustomAi(org.bukkit.entity.Entity ent, boolean init) {
      if (ent instanceof CraftCreature) {
         EntityCreature mob = ((CraftCreature)ent).getHandle();
         if (mob.bU.b().stream().noneMatch((pgw) -> {
            return pgw.i() == Integer.MAX_VALUE;
         })) {
            mob.collides = this.collides;
            mob.e(this.silent);
            mob.bU.a(Integer.MAX_VALUE, optimized);
            synchronized(mob.bT.b()) {
               if (this.typeAenabled) {
                  mob.bT.b().removeIf((pgw) -> {
                     return this.checkWrapper(pgw.k()) && StringUtils.contains(pgw.k().getClass().getName(), "ai.goal") && this.whitelistSet.stream().noneMatch((str) -> {
                        return StringUtils.contains(pgw.k().getClass().getSimpleName(), str);
                     });
                  });
               } else if (this.typeBenabled) {
                  mob.bT.b().removeIf((pgw) -> {
                     return this.checkWrapper(pgw.k()) && this.blacklistSet.stream().anyMatch((str) -> {
                        return StringUtils.contains(pgw.k().getClass().getSimpleName(), str);
                     });
                  });
               } else if (this.typeCenabled) {
                  mob.bT.b().removeIf((pgw) -> {
                     return this.checkWrapper(pgw.k());
                  });
               }

               if (mob instanceof EntityAnimal) {
                  if (this.breedEnabled) {
                     this.overriteGoal(mob.bT, PathfinderGoalBreed.class, 0, new OptimizedBreedGoal((EntityAnimal)mob));
                  }

                  if (this.panicEnabled) {
                     this.overriteGoal(mob.bT, PathfinderGoalPanic.class, 2, new OptimizedPanicGoal(mob));
                  }
               }

               if (!this.temptEnabled) {
                  return;
               }

               TagKey<Item> item = (TagKey)items.get(mob.getClass());
               if (item != null) {
                  this.overriteGoal(mob.bT, PathfinderGoalTempt.class, 1, new OptimizedTemptGoal(mob, item));
               }
            }
         }
      }

   }

   public void overriteGoal(PathfinderGoalSelector goalSelector, Class<?> clazz, int piority, PathfinderGoal goal) {
      goalSelector.b().removeIf((wg) -> {
         return wg.k().getClass().equals(clazz);
      });
      goalSelector.a(piority, goal);
   }

   public boolean checkWrapper(PathfinderGoal goal) {
      return !goal.getClass().getName().contains("mobchip") || !this.mobchip;
   }

   static {
      items.put(EntityHorse.class, TagsItem.ak);
      items.put(EntityCow.class, TagsItem.af);
      items.put(EntitySheep.class, TagsItem.ah);
      items.put(EntityFox.class, TagsItem.ae);
      items.put(EntityPig.class, TagsItem.ax);
      items.put(EntityChicken.class, TagsItem.ap);
      items.put(EntityParrot.class, TagsItem.aC);
      items.put(Frog.class, TagsItem.aq);
      items.put(Axolotl.class, TagsItem.aE);
      items.put(Goat.class, TagsItem.ag);
      items.put(EntityBee.class, TagsItem.ao);
      items.put(EntityWolf.class, TagsItem.ai);
      items.put(EntityTurtle.class, TagsItem.aB);
      items.put(EntityStrider.class, TagsItem.az);
      items.put(EntityRabbit.class, TagsItem.ay);
      items.put(EntityPiglin.class, TagsItem.ad);
      items.put(EntityPanda.class, TagsItem.av);
      items.put(EntityOcelot.class, TagsItem.au);
      items.put(EntityLlama.class, TagsItem.as);
      items.put(EntityHoglin.class, TagsItem.ar);
      items.put(Camel.class, TagsItem.am);
      items.put(Armadillo.class, TagsItem.an);
      items.put(Sniffer.class, TagsItem.ac);
   }

   public class OptimizedBreedGoal extends PathfinderGoal {
      protected final EntityAnimal animal;
      protected EntityAnimal partner;

      public OptimizedBreedGoal(EntityAnimal entityanimal) {
         this.animal = entityanimal;
         this.a(EnumSet.of(Type.a));
      }

      public boolean b() {
         if (CustomAi.this.breedEvent) {
            EntityTargetLivingEntityEvent event = CraftEventFactory.callEntityTargetLivingEvent(this.animal, this.partner, TargetReason.CUSTOM);
            if (event.isCancelled()) {
               return false;
            }
         }

         return this.animal.gy() && (this.partner = this.getFreePartner()) != null;
      }

      public boolean c() {
         return this.partner.bL() && this.partner.gy();
      }

      public void a() {
         if (CustomAi.this.breedTeleport) {
            this.animal.c(this.partner.dB(), this.partner.dD(), this.partner.dH());
         } else {
            this.animal.L().a(this.partner, CustomAi.this.breedSpeed);
         }

         this.animal.a(this.animal.dW().getMinecraftWorld(), this.partner);
      }

      private EntityAnimal getFreePartner() {
         List<? extends EntityAnimal> nearbyEntities = a(this.animal).a(this.animal.getClass(), CustomAi.this.breedTargeting, this.animal, this.animal.cR().g(8.0));
         EntityAnimal var10000;
         if (nearbyEntities.isEmpty()) {
            var10000 = null;
         } else {
            Stream var2 = nearbyEntities.stream();
            EntityAnimal var10001 = this.animal;
            Objects.requireNonNull(var10001);
            var2 = var2.filter(var10001::a);
            var10001 = this.animal;
            Objects.requireNonNull(var10001);
            var10000 = (EntityAnimal)var2.min(Comparator.comparingDouble(var10001::g)).orElse((Object)null);
         }

         return var10000;
      }
   }

   public class OptimizedPanicGoal extends PathfinderGoal {
      protected final EntityCreature mob;
      private EntityLiving lastHurtByMob;
      private int cooldown = 0;

      public OptimizedPanicGoal(EntityCreature mob) {
         this.mob = mob;
         this.a(EnumSet.of(Type.a));
      }

      public boolean b() {
         return this.cooldown-- <= 0 && (this.lastHurtByMob = this.mob.eq()) != null;
      }

      public boolean c() {
         return this.mob.g(this.lastHurtByMob) < CustomAi.this.panicRange;
      }

      public void a() {
         Vec3D randomPos = DefaultRandomPos.a(this.mob, 16, 7, this.lastHurtByMob.du());
         if (randomPos != null) {
            this.cooldown = CustomAi.this.panicCooldown;
            this.mob.L().a(randomPos.d, randomPos.e, randomPos.f, CustomAi.this.panicSpeed);
         }

      }
   }

   public class OptimizedTemptGoal extends PathfinderGoal {
      private final EntityCreature mob;
      private final PathfinderTargetCondition targetingConditions;
      private final TagKey<Item> item;
      private int cooldown = 0;

      public OptimizedTemptGoal(EntityCreature mob, TagKey<Item> item) {
         this.mob = mob;
         this.item = item;
         this.a(EnumSet.of(Type.a));
         this.targetingConditions = CustomAi.this.temptTargeting.c().a(this.mob.h(GenericAttributes.E)).a(this::shouldFollow);
      }

      public boolean b() {
         return this.cooldown-- <= 0;
      }

      private boolean shouldFollow(EntityLiving entity, WorldServer level) {
         return CustomAi.this.temptTriggerBothHands ? entity.eZ().a(this.item) || entity.fa().a(this.item) : entity.eZ().a(this.item);
      }

      public void a() {
         this.cooldown = CustomAi.this.temptCooldown;
         EntityHuman player = a(this.mob).a(this.targetingConditions, this.mob);
         if (player != null) {
            if (CustomAi.this.temptEvent) {
               EntityTargetLivingEntityEvent event = CraftEventFactory.callEntityTargetLivingEvent(this.mob, player, TargetReason.TEMPT);
               if (event.isCancelled()) {
                  return;
               }
            }

            if (this.mob.g(player) < 6.25 && !CustomAi.this.temptTeleport) {
               this.mob.L().o();
            } else if (CustomAi.this.temptTeleport) {
               this.mob.c(player.dB(), player.dD(), player.dH());
            } else {
               this.mob.L().a(player, this.mob instanceof EntityAnimal ? CustomAi.this.temptSpeed : 0.35);
            }

         }
      }
   }
}
